#
# kupo!
#
# moogle 1.0c
# -----------
# this is a say2 console-mode client written in python.  it uses asyncore and
# the say2_client module to send/receive say2 protocol messages, and some
# custom ncurses widgets to display a console similar to the one in ircII/epic
# or tf.
#
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public
# License along with this program; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.
#
# (c) 2001  Robey Pointer  <robey@lag.net>
#


import sys, os

# if 'say2cmd.py' is in this dir, add it to the path
if os.access(os.path.abspath('./say2cmd.py'), os.R_OK):
    sys.path.append(os.path.abspath('.'))
# also check for a ../lib dir that might hold our goodies
if os.access(os.path.abspath('./../lib'), os.R_OK|os.X_OK):
    sys.path.append(os.path.abspath('./../lib'))

import curses, asyncore, select, string, socket, errno, time, types, re

# okay, these all come from our own dirs:
try:
    import localizer
    from Console import Console, make_printable
    from say2_client import Say2Client
    from enaml_client import compact_traceback
    from say2set import *
    from say2cmd import *
except ImportError, e:
    print
    print 'Failed to import: ' + str(e)
    print 'This is a package that comes with moogle -- probably you need'
    print 'to add the proper path to your $PYTHONPATH.'
    print
    sys.exit(1)


class ClientEngine:

    CLIENT_NAME = 'moogle 1.0c [nov01]'
    ENCODING = 'iso-8859-1'
    NICK = ''
    ID = ''
    TIMESTAMP = 1
    BEEP_ON_MSG = 1
    MORE_MODE = 1
    HIGHLIGHT_NICK = 1
    reverse = '\006V+\006'
    unreverse = '\006V-\006'

    cmd_table = {
        # cmd: [ minimum-params, maximum-params, func, usage-str ]
        'act': [ 2, 2, cmd_act, '<nick> <action...>' ],
        'allow': [ 1, 2, cmd_allow, '<id> [message...]' ],
        'attach': [ 2, 2, cmd_attach, '<id> <password>' ],
        'away': [ 0, 1, cmd_away, '[reason...]' ],
        'bind': [ 0, 2, cmd_bind, '[key [action...]]' ],
        'boot': [ 1, 2, cmd_boot, '<nick> [reason...]' ],
        'buddies': [ 0, 0, cmd_buddies, '' ],
        'busy': [ 0, 1, cmd_busy, '[reason...]' ],
        'clear': [ 0, 0, cmd_clear, '' ],
        'cork': [ 1, 1, cmd_cork, '<on/off>' ],
        'deny': [ 1, 2, cmd_deny, '<id> [message...]' ],
        'eval': [ 1, 1, cmd_eval, '<python-expression...>' ],
	'exit': 'quit',
        'guest': [ 2, 3, cmd_guest, '<nick> <email> [tagline...]' ],
        'help': [ 0, 0, cmd_help, '' ],
        'here': [ 0, 1, cmd_here, '[reason...]' ],
        'hist': [ 0, 1, cmd_hist, '[max-lines]' ],
        'login': [ 2, 2, cmd_login, '<id> <password>' ],
        'logout': [ 0, 0, cmd_logout, '' ],
        'm': 'msg',
        'me': [ 1, 1, cmd_me, '<action...>' ],
        'motd': [ 0, 0, cmd_motd, '' ],
        'msg': [ 2, 2, cmd_msg, '<nick> <message...>' ],
	'nick': [ 1, 1, cmd_nick, '<new-nick>' ],
	'note': [ 2, 2, cmd_note, '<id> <message...>' ],
        'passwd': [ 1, 1, cmd_passwd, '<new-password>' ],
        'pref': [ 2, 2, cmd_pref, '<preference> <value>' ],
        'query': [ 0, 1, cmd_query, '[nick]' ],
        'quit': [ 0, 0, cmd_quit, '' ],
        'raw': [ 1, 1, cmd_raw, '<say2-protocol...>' ],
        'seen': [ 1, 1, cmd_seen, '<id>' ],
        'server': [ 1, 2, cmd_server, '<host> [port]' ],
        'set': [ 0, 2, cmd_set, '[pref [value...]]' ],
        'subscribe': [ 1, 2, cmd_subscribe, '<id> [message...]' ],
        'unsubscribe': [ 1, 1, cmd_unsubscribe, '<id>' ],
        'version': [ 0, 0, cmd_version, '' ],
        'wall': [ 1, 1, cmd_wall, '<message...>' ],
        'who': [ 0, 1, cmd_who, '[grep]' ],
        'whois': [ 1, 1, cmd_whois, '<nick>' ],
        'whom': [ 0, 1, cmd_whom, '[grep]' ],
        'whowas': [ 1, 1, cmd_whowas, '<id>' ],
    }

    CONFIG = {
        'encoding': 'iso-8859-1',
        'timestamp': 'on',
        'indent': '4',
        'word-wrap': 'on',
        'status-bar': '%(more)s moogle : %(nick)s %(mode)s\t%(quo)s %(ovr)s  %(time)s ',
        'more-mode': 'on',
        'beep-on-msg': 'on',
        'highlight-nick': 'on',
        'entry-size': '2,5',
        'cork-on-away': 'off',
        'idle-limit': '10',
    }

    CONFIG_TABLE = {
        'encoding': [ set_encoding, 'character encoding (Americans leave this alone)' ],
        'timestamp': [ set_timestamp, 'show timestamps on all messages/chatter (on/off)' ],
        'indent': [ set_indent, 'how much to indent wrapped lines on message/chatter' ],
        'word-wrap': [ set_word_wrap, 'wrap lines at word breaks (on/off)' ],
        'status-bar': [ set_status_bar, 'string to display on the status bar' ],
        'more-mode': [ set_more_mode, 'give [MORE] prompt when the screen fills up' ],
        'beep-on-msg': [ set_beep_on_msg, 'beep when you receive a private message' ],
        'highlight-nick': [ set_highlight_nick, 'highlight chat lines with your nick in them' ],
        'entry-size': [ set_entry_size, 'the min and max # of lines to use for the text entry field' ],
        'cork-on-away': [ set_cork_on_away, 'block incoming messages when you are /away' ],
        'idle-limit': [ set_idle_limit, 'minutes of inactivity before you are marked as idle (0 = off)' ],
    }


    def __init__(self, console, say2):
        self.console = console
        self.say2 = say2
        console.input_signal = self.do_input
        self.welcomed = 0
        self.utf8 = localizer.Localizer(self.ENCODING)
        self.console.set_more_mode(self.MORE_MODE)
        self.history_mode = 0
        self.history_mode_ancient = 0
        self.whom_mode = 0
        self.whom_list = [ ]
        self.away = 0
        self.idle = 0
        self.query = ''
        self.update_status_line = 1
        self.recently_set_icon = 0
        self.activity = int(time.time())
        self.quote_ctrl = 1
        # map of id->mode for status messages
        self.last_seen_mode = { }
        # set all the default prefs
        for i in self.CONFIG.keys():
            if self.CONFIG_TABLE.has_key(i):
                self.CONFIG_TABLE[i][0](self, self.CONFIG[i])

    def _timestr(self, epoch):
        return time.strftime('%H:%M:%S', time.localtime(epoch))
    def _timestr_ancient(self, epoch):
        return time.strftime('%d %b %H:%M:%S', time.localtime(epoch))

    def timestr(self, epoch):
        if not self.TIMESTAMP:
            return ''
        if self.history_mode_ancient:
            return '[' + self._timestr_ancient(epoch) + '] '
        return '[' + self._timestr(epoch) + '] '

    def timespan_str(self, sec):
        "timespan_str(305) = '5 min, 5 sec'"
        day, hr, min = 0, 0, 0
        if sec > 59:
            min = int(sec/60)
            sec = sec % 60
        if min > 59:
            hr = int(min/60)
            min = min % 60
        if hr > 23:
            day = int(hr/24)
            hr = hr % 24
        out = ''
        if day > 1:
            out = out + '%d days' % (day,)
        if day == 1:
            out = out + '%d day' % (day,)
        if hr > 0:
            if len(out) > 0: out = out + ', '
            out = out + '%d hr' % (hr,)
        if (min > 0) and (day < 1):
            if len(out) > 0: out = out + ', '
            out = out + '%d min' % (min,)
        if len(out) == 0:
            out = '<1 min'
        return out

    def set_quote_ctrl(self, q):
        self.quote_ctrl = q
        self.console.set_quote_ctrl(q)
        
    def highlight_nick(self, text):
        if not self.HIGHLIGHT_NICK:
            return text
        text = re.sub(r'(?i)\b(' + self.NICK + r')\b',
                      self.reverse + r'\1' + self.unreverse, text)
        return text

    def quote_printable(self, text):
        text = make_printable(text, self.quote_ctrl)
        if string.find(text, ' ') >= 0:
            text = '"' + text + '"'
        return text

    def show_whom(self):
        "dump built-up whom info"

        # find longest nick & email
        max_nick, max_email = 0, 0
        for x in self.whom_list:
            if not x.has_key('nick'): x['nick'] = ''
            if not x.has_key('email'): x['email'] = ''
            if not x.has_key('mode'): x['mode'] = '?'
            if len(x['nick']) > max_nick:
                max_nick = len(x['nick'])
            if len(x['email']) > max_email:
                max_email = len(x['email'])

        # dump.
        for x in self.whom_list:
            out = '***     %-*s <%s>%*s  ' % (max_nick, x['nick'], x['email'], max_email - len(x['email']), '')
            if x.has_key('login') and (int(x['login']) > 0):
                if time.time() - int(x['login']) > 43200:
                    out = out + time.strftime('login @ %d-%b ', time.localtime(x['login']))
                else:
                    out = out + time.strftime('login @ %H:%M  ', time.localtime(x['login']))
            mode = string.lower(x['mode'])
            if mode == 'idle':
                if x.has_key('status-time'):
                    out = out + '[idle ' + self.timespan_str(int(x['status-time'])) + '] '
                else:
                    out = out + '[idle] '
            elif (mode == 'away') or (mode == 'busy'):
                stat = string.upper(mode)
                if x.has_key('status-time'):
                    out = out + '[' + stat + ' ' + self.timespan_str(x['status-time']) + '] '
                else:
                    out = out + '[' + stat + '] '
            self.console.add(out)

            if x.has_key('status') and (len(x['status']) > 0):
                if mode == 'active': mode = 'here'
                self.console.add('***         ' + mode + ': ' + self.utf8.decode(x['status']))

        return
        
    def event_hello(self, version):
	self.protocol_version = version

    def event_server_info(self, server, version, text, uptime):
	pver0, pver1 = self.protocol_version
	self.console.add('*** say2 server (%s/%s) [protocol %d.%d]' %
	                 (self.utf8.decode(server), version, pver0, pver1))
	if uptime > 0:
	    self.console.add('*** Online since %s (%s)' %
                             (time.strftime('%H:%M %d-%b-%Y', time.localtime(uptime)),
			      self.timespan_str(time.time() - uptime)))
        if len(text):
	    self.console.add('*** ' + self.utf8.decode(text))

    def event_server_cert(self, host, cert):
        self.console.add('*** Got server cert for %s, but dunno what to do with it.' % (host,))
        
    def event_motd(self, text):
        self.console.add('*** [MOTD] ' + self.utf8.decode(text))

    def event_bad_nick(self, nick, text):
        self.console.add('*** Bad nickname ("%s"): %s' % (nick, text))

    def event_bad_address(self, address):
        self.console.add('*** Bad address ("%s")' % (address,))

    def event_logout(self, text, detached):
        if len(detached) > 0:
            self.console.add('*** Session attached from %s' % (detached,))
        if len(text) == 0:
            text = 'Goodbye!'
        self.console.add('*** ' + self.utf8.decode(text))
        
    def event_log(self, string):
        "stupid asyncore logging"
        self.console.add('--> [ASYNCORE] ' + string)

    def event_close(self):
        self.console.add('*** Lost connection to server.')
        self.welcomed = 0

    def event_connect_failed(self, why):
        self.console.add('*** Failed to connect to %s:%d : %s' % (self.say2.host, self.say2.port, why[1]))

    def event_connect(self):
        self.console.add('*** Connected to %s:%d' % (self.say2.host, self.say2.port))

    def event_error(self, command, text):
        if command and text:
            self.console.add('*** Error (%s): %s' % (command, text))
        elif command:
            self.console.add('*** Error (%s)' % (command,))
        elif text:
            self.console.add('*** Error: %s' % (text,))
        else:
            self.console.add('*** Error [???]')

    def event_not_present(self, nick, id):
        if nick:
            self.console.add('*** %s is not online.' % (nick,))
        else:
            self.console.add('*** %s (id) is not online.' % (id,))

    def event_no_such_client(self, type, id):
        self.console.add('*** No such %s: %s' % (type, id))
        
    def event_welcome(self, nick, id, attached, replicant, host):
        if self.welcomed:
            # it's a replicant
            self.console.add('*** Session attached from %s' % host)
            return
        self.welcomed = 1
        if nick != id:
            self.console.add('*** Welcome, %s!  (id: %s)' % (nick, id))
        else:
            self.console.add('*** Welcome, %s!' % (id,))
        self.NICK = nick
        self.ID = id
        self.update_status_line = 1
            
        if attached:
            if replicant:
                self.console.add('*** (Attached as replicant to existing session.)')
            else:
                self.console.add('*** (Attached to existing session.)')
            # status message is now automatically sent by the server

    def event_replicant_terminated(self, host):
        self.console.add('*** Replicant from %s terminated.' % (host,))
        
    def event_signon(self, nick, id, time, email):
        self.console.add('%s*** Signon: %s <%s>' % (self.timestr(time), nick, email))

    def event_signoff(self, nick, id, time):
        self.console.add('%s*** Signoff: %s' % (self.timestr(time), nick))

    def event_change_nick(self, old_nick, id, nick):
	if old_nick == self.NICK:
	    self.console.add('*** You have changed your nick to: %s' % (nick,))
	    self.NICK = nick
            self.update_status_line = 1
	else:
	    self.console.add('*** %s is now known as %s' % (old_nick, nick))

    def event_prefs(self, tagline, notify, pwchanged, icon_set):
        if pwchanged:
            self.console.add('*** Password changed.')
        if self.recently_set_icon and icon_set:
            self.console.add('*** Icon set!')
        elif tagline != None:
            self.console.add('*** Tagline: ' + self.utf8.decode(tagline))
        # ignore notify

    def event_status(self, nick, id, time, mode, text, messages):
        if string.lower(id) == string.lower(self.ID):
            verb = 'You are now'
	    if mode == 'busy' or mode == 'away':
		self.away = 1
	    else:
		self.away = 0
            self.update_status_line = 1
        else:
            if string.lower(id) == string.lower(nick):
                verb = nick + ' is now'
            else:
                verb = '%s (%s) is now' % (nick, id)
        if len(text) > 0:
            text = ': ' + text
        else:
            text = '.'
        mode_verb = '??'
        if mode == 'active':
            mode_verb = 'online'
        elif mode == 'busy':
            mode_verb = 'busy'
        elif mode == 'away':
            mode_verb = 'away'
        elif mode == 'idle':
            mode_verb = 'idle'
        elif mode == 'offline':
            mode_verb = 'offline'
        if not messages:
            mode_verb = mode_verb + ' (msgs off)'
        # don't display if it was a change from active->idle or back
        last_mode = '*'
        if self.last_seen_mode.has_key(string.lower(id)):
            last_mode = self.last_seen_mode[string.lower(id)]
        if ((mode == 'active') and (last_mode == 'idle')) or \
           ((mode == 'idle') and (last_mode == 'active')) or \
           (mode == last_mode):
            pass
        else:
            self.console.add('%s*** %s %s%s' % (self.timestr(time), verb, mode_verb, self.utf8.decode(text)))
        self.last_seen_mode[string.lower(id)] = mode

    def event_client_info(self, block):
        if self.whom_mode:
            # save it for later
            self.whom_list.append(block)
            return

        if block.has_key('nick'):
            # online user
            flagstr = ''
            if block.has_key('guest'):
                flagstr = flagstr + '  [guest]'
            elif string.lower(block['id']) != string.lower(block['nick']):
                flagstr = flagstr + '  (normally: ' + block['id'] + ')'
            if block.has_key('mode') and (string.lower(block['mode']) == 'idle'):
                flagstr = flagstr + '  [IDLE]'
            self.console.add('*** %s <%s>%s' % (block['nick'], block['email'], flagstr))
        else:
            # offline user
            self.console.add('*** %s <%s>  [OFFLINE]' % (block['id'], block['email']))
        if block.has_key('host'):
            self.console.add('***     Connected from ' + block['host'])
        if block.has_key('tagline'):
            self.console.add('***     Tagline: ' + self.utf8.decode(block['tagline']))
        if block.has_key('client'):
            self.console.add('***     Client: ' + self.utf8.decode(block['client']))
        if block.has_key('login'):
            self.console.add('***     Login at: ' + time.strftime('%d-%b-%Y %H:%M:%S',
                                                                  time.localtime(int(block['login']))))
        if block.has_key('mode') and (string.lower(block['mode']) == 'idle') and block.has_key('status-time'):
            self.console.add('***     Idle ' + self.timespan_str(int(block['status-time'])))
        if block.has_key('mode'):
            mode = string.lower(block['mode'])
            extra_info = ''
            if block.has_key('status-time'):
                extra_info = extra_info + ' (' + self.timespan_str(int(block['status-time'])) + ')'
            if block.has_key('status'):
                extra_info = ': ' + self.utf8.decode(block['status'])
            if block.has_key('messages'):
                if block['messages'] != 'on':
                    extra_info = ' (msgs off)' + extra_info
            if mode == 'away':
                self.console.add('***     AWAY' + extra_info)
            elif mode == 'busy':
                self.console.add('***     BUSY' + extra_info)
            elif block.has_key('status'):
                self.console.add('***     %s is %s' % (block['nick'], self.utf8.decode(block['status'])))
        if block.has_key('last-logout'):
            span = time.time() - int(block['last-logout'])
            tstr = time.strftime('%d-%b-%Y %H:%M:%S', time.localtime(int(block['last-logout'])))
            if int(block['last-logout']) > 0:
                self.console.add('***     Last logged out %s (%s ago)' % (tstr, self.timespan_str(span)))
            else:
                self.console.add('***     Never logged in.')
        return

    def event_members(self, type, list):
        if type == 'user':
            self.console.add('*** Users: ' + string.join(list, ' '))
        else:
            self.console.add('*** Channels: ' + string.join(list, ' '))

    def event_say(self, nick, id, epoch, block):
        if (time.time() - epoch > 43200) and (self.history_mode):
            self.history_mode_ancient = 1
        if block.has_key('text'):
            self.console.add('%s<%s> %s' % (self.timestr(epoch), nick,
                                            self.highlight_nick(self.utf8.decode(str(block['text'])))))
        if block.has_key('act'):
            self.console.add('%s* %s %s' % (self.timestr(epoch), nick,
                                            self.highlight_nick(self.utf8.decode(str(block['act'])))))

    def event_wall(self, nick, id, epoch, block):
        if block.has_key('text'):
            self.console.add('%s*** WALL from %s: %s' % (self.timestr(epoch), nick,
                                                         self.utf8.decode(block['text'])))
        else:
            self.console.add('%s*** WALL from %s [???]' % (self.timestr(epoch), nick))
            
    def event_message(self, nick, id, to_nick, to_id, epoch, block, stored):
        if stored:
            stored_text = '(stored) '
        else:
            stored_text = ''
        # we do this as two if's instead of one if/else, because it's possible
        # that this is a message from us, to us.  in that case, the server only
        # sends one copy of the message, and it's up to us to display it twice
        # if we so desire (and we do).
        if string.lower(nick) == string.lower(self.NICK):
            # outgoing message
            if block.has_key('act'):
                self.console.add('%s%s-> [%s] * %s %s' % (self.timestr(epoch), stored_text, to_nick, self.NICK,
                                                          self.utf8.decode(str(block['act']))))
            elif block.has_key('text'):
                self.console.add('%s%s-> [%s] %s' % (self.timestr(epoch), stored_text, to_nick,
                                                     self.utf8.decode(str(block['text']))))
            else:
                self.console.add('%s%s-> [%s] (unknown message type)' % (self.timestr(epoch), stored_text, to_nick))
        if string.lower(to_nick) == string.lower(self.NICK):
            # message to me!
            if self.BEEP_ON_MSG:
                curses.beep()
            if block.has_key('act'):
                self.console.add('%s%s* [%s%s%s] %s' % (self.timestr(epoch), stored_text, self.reverse, nick,
                                                        self.unreverse, self.utf8.decode(str(block['act']))))
            elif block.has_key('text'):
                self.console.add('%s%s[%s%s%s] %s' % ( self.timestr(epoch), stored_text, self.reverse, nick,
                                                       self.unreverse, self.utf8.decode(str(block['text']))))
            else:
                self.console.add('%s%s[%s%s%s] (unknown message type)' % (self.timestr(epoch), stored_text,
                                                                          self.reverse, nick, self.unreverse))

    def event_boot_user(self, nick, id, target_nick, target_id, text):
        if text != '':
            self.console.add('*** %s booted %s: %s' % (nick, target_nick, self.utf8.decode(text)))
        else:
            self.console.add('*** %s booted %s' % (nick, target_nick))

    def event_subscriptions_from(self, approved, blocked, pending):
        blocked = map(lambda x: x+'(blocked)', blocked)
        pending = map(lambda x: x+'(pending)', pending)
        approved = approved + blocked + pending
        if len(approved) > 0:
            self.console.add('*** People who have you on their buddy list: ' + string.join(approved, ' '))

    def event_subscriptions(self, approved, blocked, pending):
        blocked = map(lambda x: x+'(blocked)', blocked)
        pending = map(lambda x: x+'(pending)', pending)
        approved = approved + blocked + pending
        if len(approved) > 0:
            self.console.add('*** Your buddy list: ' + string.join(approved, ' '))
        else:
            self.console.add('*** Your buddy list is empty.')
            
    def event_subscription_request(self, id, text):
        if len(text) > 0:
            self.console.add('*** %s wants to be your buddy: %s' % (id, self.utf8.decode(text)))
        else:
            self.console.add('*** %s wants to be your buddy.' % (id,))
        self.console.add('*** (Use "/allow <id> [message...]" to allow, or "/deny <id> [message...] to deny.)')

    def event_no_subscription_request(self, id):
        self.console.add('*** No buddy-list request from %s.' % (id,))

    def event_no_subscription(self, id):
        self.console.add('*** %s is not on your buddy list.' % (id,))

    def event_subscription_accepted(self, id, text, already):
        if len(text) > 0:
            text = ': ' + text
        else:
            text = '.'
        if already:
            self.console.add('*** %s is already your buddy.' % (id,))
        else:
            self.console.add('*** %s agrees to be your buddy%s' % (id, self.utf8.decode(text)))

    def event_subscription_denied(self, id, text, forever):
        if len(text) > 0:
            text = ': ' + text
        else:
            text = '.'
        if forever:
            self.console.add('*** You\'ve tried too many times to be buddies with %s.' % (id,))
        else:
            self.console.add('*** %s doesn\'t want to be your buddy%s' % (id, self.utf8.decode(text)))

    def event_unsubscribed_from(self, id):
        self.console.add('*** You are no longer on %s\'s buddy list.' % (id,))
        
    def event_begin_response(self, response, count, total):
        if response == 'client-info':
            self.whom_mode = 1
            self.whom_list = [ ];
            if count < total:
                self.console.add('*** %d users online (%d shown):' % (total, count))
            else:
                plur = 's'
                if count == 1: plur = ''
                self.console.add('*** %d user%s online:' % (count, plur))
            return
        if response == 'say':
            # history
            self.history_mode = 1
            self.history_mode_ancient = 0;
            self.console.add('*** History begins:')
            return
        self.console.add('*** Mystery response block (%d/%d) of type "%s":' % (count, total, response))

    def event_end_response(self, response):
        if response == 'client-info':
            self.whom_mode = 0
            self.show_whom()
            return
        if response == 'say':
            self.history_mode = 0
            self.history_mode_ancient = 0
            self.console.add('*** End of history.')
            return
        self.console.add('*** (End of mystery response block "%s")' % (response,))

    ### TextEntry input callbacks
        
    def do_input(self, text):
        "return 1 if we should quit"
        if len(text) == 0:
            return 0
        if (text[0] == '/') and not (string.lower(text[:4]) == '/say'):
            return self.do_command(text[1:])
        if not say2.connected:
            self.console.add('*** Talking is futile.  You need to connect to a server (/server).')
            return 0
        if self.away:
            self.away = 0
            if self.CORK_ON_AWAY:
                self.say2.set_status('active', '', 0)
            else:
                self.say2.set_status('active', '')
        if (not self.query) or (string.lower(text[:4]) == '/say'):
            if string.lower(text[:4]) == '/say':
                text = text[5:]
            self.say2.say(self.utf8.encode(text))
        else:
            self.say2.message(self.query, None, self.utf8.encode(text))

    def do_command(self, cmd):
        n = string.split(string.strip(cmd), ' ', 1)
        cmd = n[0]
        if len(n) > 1:
            arg = string.strip(n[1])
        else:
            arg = ''
        
        ok, cmd, answer = match_command(self.cmd_table, cmd)
        while ok and (type(answer) == types.StringType):
            ok, cmd, answer2 = match_command(self.cmd_table, answer)
            if not ok:
                self.console.add('*** Alias /%s goes to nonexistent "/%s"' % (cmd, answer))
                return 0
            answer = answer2
            
        if not ok:
            if len(answer) > 0:
                self.console.add('*** Which did you mean: ' + string.join(map(lambda x: '/'+x, answer)))
            else:
                possible = self.cmd_table.keys()
                possible.sort()
                self.console.add('*** Try: ' + string.join(map(lambda x: '/'+x, possible)))
            return 0
        if answer[1] > 1:
            arglist = string.split(arg, ' ', answer[1]-1)
            arglist = filter(lambda x: x != '', arglist)
        else:
            if arg != '':
                arglist = [ arg ]
            else:
                arglist = [ ]
        if len(arglist) < answer[0]:
            self.console.add('*** Usage: /%s %s' % (cmd, answer[3]))
            return 0

        return apply(answer[2], (self, arglist))
    
    
def match_command(table, word):
    if table.has_key(word):
        return (1, word, table[word])
    # check for closest match
    cmds = table.keys()
    cmds.sort()
    matches = [ ]
    for x in cmds:
        if (len(x) > len(word)) and (word == x[:len(word)]):
            matches.append(x)
    if len(matches) == 1:
        return (1, matches[0], table[matches[0]])
    return (0, word, matches)

def get_homedir():
    import os, pwd
    if os.environ.has_key('HOME'):
        return os.environ['HOME']
    return pwd.getpwuid(os.getuid())[5]



HOST = 'say2.lag.net'
PORT = 3797
DEBUG = 0

if (len(sys.argv) > 1) and (sys.argv[1] == '-d'):
    DEBUG = 1
    sys.argv.pop(0)
    
if len(sys.argv) > 1:
    x = string.split(sys.argv[1], ':')
    HOST = x[0]
    if len(x) > 1:
        PORT = int(x[1])
    
try:
    stdscr = curses.initscr()
    curses.noecho()
    curses.raw()
    stdscr.keypad(1)
    stdscr.nodelay(1)
    stdscr.clear()
    last_sb_update = 0
    
    console = Console(stdscr)
    console.status_table['nick'] = ''
    console.status_table['time'] = ''
    console.status_table['date'] = ''
    console.status_table['mode'] = ''
    say2 = Say2Client(HOST, PORT)
    say2.DEBUG = DEBUG
    client = ClientEngine(console, say2)
    say2.add_listener(client)

    console.add('*** \006B+\006%s\006B-\006  Kupo!' % (client.CLIENT_NAME,))
    
    # read .say2rc
    try:
        fd = open(get_homedir() + '/.say2rc', 'r')
        line = fd.readline()
        while line != '':
            line = string.strip(fd.readline())
            if len(line) == 0:
                continue
            if (line[0] == '#') or (line[0] == ';'):
                continue
            client.do_input(line)
        fd.close()
    except IOError:
        pass
    
    while console.active:
        if client.update_status_line or (last_sb_update + 15 <= time.time()):
            # update the status bar, but only every 15 seconds at most
            last_sb_update = time.time()
            lt = time.localtime(last_sb_update)
            console.status_table['nick'] = client.NICK
            console.status_table['time'] = time.strftime('%H:%M', lt)
            console.status_table['date'] = time.strftime('%d-%b', lt)
            if client.idle:
                console.status_table['mode'] = '(idle)'
            elif client.away:
                console.status_table['mode'] = '(away)'
            else:
                console.status_table['mode'] = ''
            console.update_status_line()
            console.draw()
            client.update_status_line = 0
        try:
            console.activity = 0
            asyncore.poll(15)
            if console.activity and console.active:
                client.activity = int(time.time())
                if client.idle:
                    # turn off idle mode
                    say2.set_status('active')
                    client.idle = 0
            else:
                if (not client.idle) and (not client.away) and (client.IDLE_LIMIT != 0) and \
                   (int(time.time()) - client.activity >= client.IDLE_LIMIT):
                    client.idle = 1
                    say2.set_status(mode='idle', duration=(int(time.time()) - client.activity))
        except select.error, why:
            if why[0] != errno.EINTR:
                raise
            # interrupted system call -> probably SIGWINCH
            console.handle_read()
        except:
            t, info = compact_traceback()
            console.add('!!! EXCEPTION: ' + info)
            time.sleep(1)

finally:
    curses.nocbreak()
    curses.noraw()
    curses.echo()
    stdscr.keypad(0)
    curses.endwin()

