#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public
# License along with this program; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.
#
# (c) 2000, 2001  Robey Pointer  <robey@lag.net>
#


import string, socket

def cmd_act(client, args):
    client.say2.message_act(args[0], None, client.utf8.encode(args[1]))
    
def cmd_allow(client, args):
    if len(args) == 1:
        client.say2.subscription_accept(args[0])
    else:
        client.say2.subscription_accept(args[0], client.utf8.encode(args[1]))

def cmd_away(client, args):
    if len(args) == 0:
        client.away = 0
        client.say2.set_status('active', '', 0)
    else:
        if client.CORK_ON_AWAY:
            client.say2.set_status('away', client.utf8.encode(args[0]), 1)
        else:
            client.say2.set_status('away', client.utf8.encode(args[0]))
        client.away = 1

def cmd_attach(client, args):
    client.say2.login(args[0], args[1], client.CLIENT_NAME, 1)

def cmd_buddies(client, args):
    client.say2.send_message('list-subscriptions')
    
def cmd_busy(client, args):
    if len(args) == 0:
        client.away = 0
        client.say2.set_status('active', '', 0)
    else:
        if client.CORK_ON_AWAY:
            client.say2.set_status('busy', client.utf8.encode(args[0]), 1)
        else:
            client.say2.set_status('busy', client.utf8.encode(args[0]))
        client.away = 1

def cmd_bind(client, args):
    if len(args) == 0:
        # show existing bindings
        client.console.show_all_key_bindings()
        return 0
    key = args[0]
    if key == '^?':
        key = chr(127)
    if (len(key) == 2) and (key[0] == '^') and (key[1] >= 'A') and (key[1] <= '_'):
        key = chr(ord(key[1])-64)
    if key == 'space':
        key = ' '
    key = string.lower(key)
    if len(args) == 1:
        client.console.show_key_binding(key)
        return 0

    # set the new key
    n = client.console.set_key_binding(key, args[1])
    if n == -1:
        client.console.add('*** Key binding removed.')
    elif n == 0:
        client.console.add('*** No such action defined.')
    else:
        client.console.show_key_binding(key)
    return 0

def cmd_boot(client, args):
    if len(args) == 1:
        client.say2.boot(args[0], None, None)
    else:
        client.say2.boot(args[0], None, client.utf8.encode(args[1]))

def cmd_clear(client, args):
    client.console.clear()

def cmd_cork(client, args):
    if args[0] == 'on':
        client.say2.set_cork(1)
    else:
        client.say2.set_cork(0)
        
def cmd_deny(client, args):
    if len(args) == 1:
        client.say2.subscription_deny(args[0])
    else:
        client.say2.subscription_deny(args[0], client.utf8.encode(args[1]))

def cmd_eval(client, args):
    vars = { 'say2': client.say2, 'console': client.console, 'client': client }
    try:
        exec args[0] in vars
    except e:
        client.console.add('*** Exception in /eval: ' + str(e))

def cmd_guest(client, args):
    if len(args) == 2:
        client.say2.login_guest(args[0], client.utf8.encode(args[1]), None, client.CLIENT_NAME)
    else:
        client.say2.login_guest(args[0], client.utf8.encode(args[1]),
                                client.utf8.encode(args[2]), client.CLIENT_NAME)

def cmd_help(client, args):
    if len(args) == 0:
        client.console.add('*** say2.py commands:')
        list = client.cmd_table.keys()
        list.sort()
        for cmd in list:
            if type(client.cmd_table[cmd]) != type(''):
                client.console.add('    /%s %s' % (cmd, client.cmd_table[cmd][3]))
    else:
        cmd = string.lower(args[0])
        if client.cmd_table.has_key(cmd):
            client.console.add('*** Usage: /%s %s' % (cmd, client.cmd_table[cmd][3]))
        else:
            client.console.add('*** I\'m not familiar with /%s.  Try /help for a list of commands.' % (cmd,))

def cmd_here(client, args):
    client.away = 0
    if len(args) == 0:
        client.say2.set_status('active', '', 0)
    else:
        client.say2.set_status('active', client.utf8.encode(args[0]), 0)

def cmd_hist(client, args):
    if len(args) == 0:
        client.say2.get_history()
    else:
        client.say2.get_history(int(args[0]))

def cmd_login(client, args):
    client.say2.login(args[0], args[1], client.CLIENT_NAME)
    
def cmd_logout(client, args):
    client.say2.logout()
    return 0

def cmd_me(client, args):
    client.say2.say_act(client.utf8.encode(args[0]))
    return 0

def cmd_motd(client, args):
    client.say2.send_message('motd')
    return 0
    
def cmd_msg(client, args):
    client.say2.message(args[0], None, client.utf8.encode(args[1]))
    return 0

def cmd_nick(client, args):
    client.say2.set_nick(args[0])
    return 0

def cmd_note(client, args):
    client.say2.message(None, args[0], client.utf8.encode(args[1]))
    return 0

def cmd_passwd(client, args):
    client.say2.set_password(args[0])
    return 0

def cmd_pref(client, args):
    if string.lower(args[0]) == 'tagline':
        client.say2.set_tagline(client.utf8.encode(args[1]))
        return 0
    elif string.lower(args[0]) == 'icon':
        try:
            fd = open(args[1], 'r')
        except IOError:
            client.console.add('*** Can\'t open icon file "%s"' % (args[1],))
            return 0
        try:
            icon_data = fd.read(4097)
            if len(icon_data) > 4096:
                client.console.add('*** Icon "%s" is too big (must be less than 4KB).' % (args[1],))
                return 0
        except IOError:
            client.console.add('*** Error reading icon "%s"' % (args[1],))
            return 0
        fd.close()
        client.console.add('*** Uploading new icon...')
        client.say2.set_icon(icon_data)
        client.recently_set_icon = 1
    else:
        client.console.add('*** Available preferences: tagline icon')
        return 0

def cmd_query(client, args):
    if len(args) == 0:
        if not client.query:
            client.console.add('*** You were not in /query with anyone.')
        else:
            client.console.add('*** Now speaking in the public channel.')
            client.query = ''
        return 0
    client.query = args[0]
    client.console.add('*** Now automatically messaging ' + args[0])

def cmd_quit(client, args):
    return 1

def cmd_raw(client, args):
    client.say2.send(args[0] + '\r\n')
    return 0
    
def cmd_seen(client, args):
    client.say2.whois(None, args[0])
    return 0

def cmd_server(client, args):
    if len(args) < 2:
        port = 3797
    else:
        port = int(args[1])
    if client.say2.connected:
        client.console.add('*** Disconnecting from %s:%d' % (client.say2.host, client.say2.port))
        client.say2.disconnect()
    try:
        client.say2.connect(args[0], port)
        # signal will trigger if connect succeeds
    except socket.error, why:
        client.console.add('*** Failed to connect to %s:%d : %s' % (args[0], port, why[1]))
    return 0

def cmd_set(client, args):
    if len(args) == 0:
        client.console.add('*** /set <pref> <value>')
        list = client.CONFIG.keys()
        list.sort()
        maxwidth = reduce(lambda n, x: max(n, len(x)), list, 0)
        for i in list:
            out = client.quote_printable(client.CONFIG[i])
            client.console.add('*** %-*s: %s' % (maxwidth, i, out))
            if client.CONFIG_TABLE.has_key(i):
                client.console.add('***     %s' % (client.CONFIG_TABLE[i][1],))
        return 0

    key = string.lower(args[0])
    if not client.CONFIG.has_key(key):
        client.console.add('*** No such setting.')
        return 0

    if len(args) == 1:
        out = client.quote_printable(client.CONFIG[key])
        client.console.add('*** %s: %s' % (key, out))
        if client.CONFIG_TABLE.has_key(key):
            client.console.add('***     %s' % (client.CONFIG_TABLE[key][1],))
        return 0

    client.CONFIG[key] = args[1]
    if client.CONFIG_TABLE.has_key(key):
        client.CONFIG_TABLE[key][0](client, args[1])
    out = client.quote_printable(client.CONFIG[key])
    client.console.add('*** Set %s: %s' % (key, out))
    return 0

def cmd_subscribe(client, args):
    if len(args) == 1:
        client.say2.subscribe(args[0])
    else:
        client.say2.subscribe(args[0], client.utf8.encode(args[1]))
    client.console.add('*** Asking to subscribe to %s' % (args[0],))
    return 0

def cmd_unsubscribe(client, args):
    client.say2.unsubscribe(args[0])
    return 0

def cmd_version(client, args):
    client.console.add('*** %s (encoding: %s)' % (client.CLIENT_NAME, client.ENCODING))
    return 0

def cmd_wall(client, args):
    client.say2.wall(args[0])
    return 0

def cmd_who(client, args):
    if len(args) > 0:
        client.say2.list_users(0, args[0])
    else:
        client.say2.list_users()
    return 0

def cmd_whois(client, args):
    client.say2.whois(args[0], None)
    return 0

def cmd_whom(client, args):
    if len(args) > 0:
        client.say2.list_users(1, args[0])
    else:
        client.say2.list_users(1)
        
def cmd_whowas(client, args):
    client.say2.whois(None, args[0])
    return 0

def cmd_zzz(client, args):
    for i in range(30):
        client.console.add('Line %d' % (i,))
    return 0
